"""
Main entrypoint for terraplanfeed command.

Main code execution for terraplanfeed command.

Functions:
    terraplanfeed: main code execution
"""
import logging
import sys
import json

from terraplanfeed.terraform import parsePlan
from terraplanfeed.stdout import generate_stdout
from terraplanfeed.azuredevops import generate_pr_comment

logger = logging.getLogger(__name__)


def terraplanfeed(filename, output):
    """
    Execute terraplanfeed.

    Args:
        filename: Terraform plan in JSON format
        output: output driver to write feedback

    Returns:
        Boolean
    """

    """Open JSON file"""
    logger.debug("reading file %s", filename)
    with open(filename) as f:
        try:
            tf = json.load(f)
        except ValueError as e:
            logger.error("%s", e)
            sys.exit(1)

    resources = parsePlan(tf)

    if output.lower() == "azuredevops":
        generate_pr_comment(resources)
    else:
        generate_stdout(resources)



"""
Entry point for package.

Arguments:
    -h, --help: prints help
    -V, --version: prints version, location and python powering it
    <filename> (required): terraform plan in json format
    -v, --verbose: prints debug information
    -d, --debug-file: filename to print debug information
    -o, --output: output driver
"""

import os
import sys
import click

from terraplanfeed import __version__
from terraplanfeed.log import configure_logger
from terraplanfeed.main import terraplanfeed


def version_msg():
    """Return the Terraplanfeed version, location and Python powering it."""
    python_version = sys.version[:3]
    location = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    message = "Terraplanfeed %(version)s from {} (Python {})"
    return message.format(location, python_version)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__, "-V", "--version", message=version_msg())
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Print debug information",
    default=False,
)
@click.option(
    "-d",
    "--debug-file",
    type=click.Path(),
    default=None,
    help="Print debug information to this file",
)
@click.option(
    "-o",
    "--output",
    type=click.Choice(["stdout", "azuredevops"], case_sensitive=False),
    default="stdout",
    help="Output driver",
)
def main(filename, verbose, debug_file, output):
    """Parse Terraform plan in JSON format."""

    configure_logger(
        stream_level="DEBUG" if verbose else "INFO", debug_file=debug_file
    )

    terraplanfeed(filename, output)


if __name__ == "__main__":
    main()



"""
Azure DevOps module for Pull Request comment with changes.

This module expects the following ENV variables:
- SYSTEM_TEAMFOUNDATIONSERVERURI
- SYSTEM_TEAMPROJECT
- BUILD_REPOSITORY_ID
- SYSTEM_PULLREQUEST_PULLREQUESTID
- SYSTEM_ACCESSTOKEN

if any of these variables is empty it will use stdout driver.

Some useful links:
https://docs.microsoft.com/en-us/azure/devops/pipelines/build/variables?view=azure-devops&tabs=yaml#system-variables-devops-services
https://docs.microsoft.com/en-us/rest/api/azure/devops/git/pull%20request%20threads/create?view=azure-devops-rest-6.0
https://docs.microsoft.com/en-us/rest/api/azure/devops/git/pull%20request%20threads/create?view=azure-devops-rest-6.0#commentthreadstatus
https://docs.microsoft.com/en-us/rest/api/azure/devops/git/pull%20request%20threads/create?view=azure-devops-rest-6.0#comment
https://docs.microsoft.com/en-us/rest/api/azure/devops/git/pull%20request%20threads/create?view=azure-devops-rest-6.0#commenttype

Functions:
    getAction: gets the action from actions list
    parseChanges: gets list of changes and creates multiline summary
    validateEnvVars: validates required env variables
    formatUrl: formats URL using environment variables
    formatDataToSend: creates data to be sent in the request
    sendRequest: send the request with the content to devops
    main: entrypoint
"""
import logging
import os
import base64
import requests
from terraplanfeed.stdout import generate_stdout

logger = logging.getLogger(__name__)

URLCOMPONENTS = {
    "URL": os.getenv("SYSTEM_TEAMFOUNDATIONSERVERURI"),
    "PROJ": os.getenv("SYSTEM_TEAMPROJECT"),
    "REPOSITORYID": os.getenv("BUILD_REPOSITORY_ID"),
    "PULLREQUESTID": os.getenv("SYSTEM_PULLREQUEST_PULLREQUESTID"),
    "APIVERSION": "6.0",
}

COMMENTTYPE = {
    "CODECHANGE": "codeChange",
    "SYSTEM": "system",
    "TEXT": "text",
    "UNKNOWN": "unknown",
}

COMMENTTHREADSTATUS = {
    "ACTIVE": "active",
    "BYDESIGN": "byDesign",
    "CLOSED": "closed",
    "FIXED": "fixed",
    "PENDING": "pending",
    "UNKNOWN": "unknown",
    "WONTFIX": "wontFix",
}

ACTION_SYMBOLS = {
    "no-op": ":thumbsup:",
    "create": ":sparkles:",
    "read": ":open_book:",
    "update": ":pencil2:",
    "replace": ":warning:",
    "delete": ":stop_sign:",
}

HEADER = """
**Terraform Plan changes summary:**
===================================
"""

FOOTER = """
"""


def getAction(actions):
    """
    Get action

    Args:
        actions(list): list of actions

    Returns:
        action symbol
    """

    logger.debug("get action")
    if "create" in actions and len(actions) > 1:
        return ACTION_SYMBOLS["replace"]
    else:
        return ACTION_SYMBOLS[actions[0]]


def parseChanges(changes):
    """
    Parse changes.

    Args:
        changes(list): list of resources dict

    Returns:
        Multiline string with summary of changes
    """

    content = ""
    logger.debug("parsing changes...")
    for c in changes:
        action = getAction(c["actions"])
        message = "({action}): {name} ({address})".format(
            action=action, name=c["name"], address=c["address"]
        )
        content += message + "\n"
    return content


def validateEnvVars():
    """
    Validates the contents of the required env variables

    Returns:
        boolean true or false if they are present
    """

    logger.debug("check required env variables")
    if URLCOMPONENTS["URL"] is None:
        return False
    if URLCOMPONENTS["PROJ"] is None:
        return False
    if URLCOMPONENTS["REPOSITORYID"] is None:
        return False
    if URLCOMPONENTS["PULLREQUESTID"] is None:
        return False
    if os.getenv("SYSTEM_ACCESSTOKEN") is None:
        return False

    return True


def formatUrl():
    """
    Formats URL using environment variables.

    Returns: url
    """

    url = ""
    logger.debug("format url")
    url = "{url}{proj}/_apis/git/repositories/{repoid}/pullRequests/{prid}/threads?api-version={apiversion}".format(
        url=URLCOMPONENTS["URL"],
        proj=URLCOMPONENTS["PROJ"],
        repoid=URLCOMPONENTS["REPOSITORYID"],
        prid=URLCOMPONENTS["PULLREQUESTID"],
        apiversion=URLCOMPONENTS["APIVERSION"],
    )

    return url


def formatDataToSend(content, commentType, statusCode):
    """
    Creates JSON content to send to Azure DevOps.

    Args:
        content(str): multiline string
        commentType(str): comment type
        statusCode(str): status code

    Returns:
        dict with the content
    """

    formattedContent = HEADER + content + FOOTER
    data = {
        "comments": [
            {
                "parentCommentId": 0,
                "content": formattedContent,
                "commentType": commentType,
            }
        ],
        "status": statusCode,
    }

    return data


def sendRequest(url, data):
    """
    Sends request to Azure DevOps API.

    Args:
        url: URL to send request
        data: json content to send to API

    Returns:
        True or False based on the send request
    """

    logger.debug("send request")
    logger.debug("url: {}".format(url))
    logger.debug("data: {}".format(data))

    token = os.getenv("SYSTEM_ACCESSTOKEN")
    access_token = ":{}".format(token)

    encodedBytes = base64.b64encode(access_token.encode("utf-8"))
    encodedStr = str(encodedBytes, "utf-8")

    session = requests.Session()
    session.headers["Authorization"] = "Basic {}".format(encodedStr)
    response = session.post(url, json=data)

    return response.status_code


def generate_pr_comment(changes):
    """
    Handles changes and formats content to send to Azure DevOps API.

    Args:
        changes(list): list of resources dict
    """

    ret = False
    logger.debug("azure devops entrypoint")
    if not changes:
        content = "No changes " + ACTION_SYMBOLS["no-op"]
        commentType = COMMENTTYPE["TEXT"]
        statusCode = COMMENTTHREADSTATUS["CLOSED"]
    else:
        content = parseChanges(changes)
        commentType = COMMENTTYPE["CODECHANGE"]
        statusCode = COMMENTTHREADSTATUS["ACTIVE"]
    data = formatDataToSend(content, commentType, statusCode)
    if validateEnvVars():
        url = formatUrl()
        retcode = sendRequest(url, data)
        logger.debug("status code: {}".format(retcode))
        if retcode == 200:
            ret = True
    else:
        generate_stdout(content)
        ret = True

    return ret



"""Terraform handler.

This module handles terraform plan in json output format.

Check https://www.terraform.io/docs/internals/json-format.html
for more info

Functions:
    getResourceChanges: filters resource_changes from plan file
    filterNoOp: filters out no-op actions from resource_changes
    parseResource: parses resource to get some metadata like,
        name, address
    calculateName: tries to calculate resource name
    getAttributes: creates a list of resources with some attributes
    parsePlan: main function
"""
import logging

logger = logging.getLogger(__name__)


def getResourceChanges(data):
    """
    Filters Resource Changes from Plan.

    Retrieves the resource_changes portion from terraform plan.

    Args:
        data (dict): dict from json.load('<terraform plan in json format>')

    Returns:
        List of resource_changes
    """

    logger.debug("filter resource_changes from plan")
    if "resource_changes" in data.keys():
        return data["resource_changes"]
    else:
        logger.error("no resource_changes found")
        return []


def filterNoOp(resources):
    """
    Filters no-op actions.

    Parses all the resource_changes and cleans out the ones with no-op actions

    Args:
        data(list): list of resource_changes (output from getResourceChanges)

    Returns:
        List of resources that are going to be changed
    """

    changes = []
    logger.debug("filter no-op actions")
    for res in resources:
        if res["change"]["actions"] != ["no-op"]:
            changes.append(res)

    return changes


def calculateName(resource):
    """
    Tries to calculate the name of the resource.

    Args:
        resource(dict): Resource object

    Returns:
        String name
    """

    name = ""
    beforeName = ""
    afterName = ""
    logger.debug("tries to calculate resource name")
    if resource["change"]["before"] is not None:
        if "name" in resource["change"]["before"]:
            beforeName = resource["change"]["before"]["name"]
            logger.debug("beforeName: %s", beforeName)
    if resource["change"]["after"] is not None:
        if "name" in resource["change"]["after"]:
            afterName = resource["change"]["after"]["name"]
            logger.debug("afterName: %s", afterName)

    if afterName and beforeName and (afterName != beforeName):
        name = beforeName + " --> " + afterName
    elif afterName and beforeName:
        name = beforeName
    elif beforeName and (afterName is not None):
        name = beforeName
    elif afterName and (beforeName is not None):
        name = afterName
    else:
        name = "<known after apply>"

    return name


def parseResource(resource):
    """
    Parse Resource to retrieve metadata.

    Args:
        resource(dict): Resource object

    Returns:
        Dict with resources attributes
    """

    logger.debug("retrieve elements from resource")
    rc = {
        "actions": resource["change"]["actions"],
        "name": calculateName(resource),
        "address": resource["address"],
    }
    return rc


def getAttributes(data):
    """
    Get attributes from resource.

    Args:
        data(list): list of resources

    Returns:
        List of resources dict with attributes
    """

    resources = []
    logger.debug("get attributes from resources")
    for res in data:
        resources.append(parseResource(res))
    return resources


def parsePlan(tfplan):
    """
    Parse Terraform plan in JSON.

    Args:
        tfplan(dict): dict from json.load('<terraform plan in json format>')

    Returns:
        List of resources dict with attributes
    """

    logger.debug("parsing file...")

    all_resources = getResourceChanges(tfplan)
    logger.debug(all_resources)

    resource_changing = filterNoOp(all_resources)
    logger.debug(resource_changing)

    resources = getAttributes(resource_changing)
    logger.debug(resources)

    return resources



"""
STDOUT output driver module.

https://www.terraform.io/docs/internals/json-format.html#change-representation

Functions:
    getAction: gets the action from actions list
    parseChanges: gets list of changes and creates multiline summary
    write: writes the summary content to stdout
    main: entrypoint
"""
import logging

logger = logging.getLogger(__name__)

ACTION_SYMBOLS = {
    "no-op": "None",
    "create": "Create",
    "read": "Read",
    "update": "Update",
    "replace": "Replace",
    "delete": "Delete",
}

HEADER = """
**Terraform Plan changes summary:**
===================================
"""

FOOTER = """
"""


def getAction(actions):
    """
    Get action

    Args:
        actions(list): list of actions

    Returns:
        action symbol
    """

    logger.debug("get action")
    if "create" in actions and len(actions) > 1:
        return ACTION_SYMBOLS["replace"]
    else:
        return ACTION_SYMBOLS[actions[0]]


def parseChanges(changes):
    """
    Parse changes.

    Args:
        changes(list): list of resources dict

    Returns:
        Multiline string with summary of changes
    """

    content = ""
    logger.debug("parsing changes...")
    for c in changes:
        action = getAction(c["actions"])
        message = "({action}): {name} ({address})".format(
            action=action, name=c["name"], address=c["address"]
        )
        content += message + "\n"
    return content


def write(content):
    """
    Writes summary of changes to stdout.

    Args:
        content(str): multiline string
    """

    logger.debug("write to stdout")
    print(HEADER)
    print(content)
    print(FOOTER)


def generate_stdout(changes):
    """
    Entrypoint for stdout output driver.

    Args:
        changes(list): list of resources dict
    """

    logger.debug("stdout entrypoint")
    if not changes:
        content = "No changes"
    else:
        content = parseChanges(changes)
    write(content)



"""Setup for Terraplanfeed."""

from setuptools import setup, find_packages
import pathlib
import terraplanfeed

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="terraplanfeed",
    version=terraplanfeed.__version__,
    description="Parse Terraform plan in json format and give feedback.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bcochofel/terraplanfeed",
    author="Bruno Cochofel",
    author_email="bruno.cochofel@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Build Tools",
    ],
    keywords=["terraplanfeed", "Python", "Terraform", "Terraform Plan"],
    packages=["terraplanfeed"],
    python_requires=">=3.7, <4",
    install_requires=["Click", "Requests"],
    entry_points={
        "console_scripts": ["terraplanfeed=terraplanfeed.__main__:main"]
    },
)



"""
Main package for terraplanfeed.

This package receives a file in Terraform JSON output, parses to get the
changes, and creates output that can be sent to different output drivers.

Modules:
    terraform
"""
__version__ = "0.6.3"
